"""Compute multi-commodity flow in grid."""

from typing import Sequence, Mapping, Optional
from z3 import Bool, SolverFor, Or, PbLe, Implies, PbEq, Not, sat # type: ignore

class GridAnalysis:
    """Class for the results of the preprocessing analyses."""
    def __init__(self, grid: Sequence[Sequence[Optional[int]]]):
        self.nrow = len(grid)
        self.ncol = len(grid[0])

        self.ngrid: list[list[Optional[int]]] = \
            [[grid[i][j] for j in range(self.ncol)] for i in range(self.nrow)]

        self.regions: list[list[set[tuple[int,int]]]] = \
            [[({(i,j,)} if self.ngrid[i][j] is not None else set())
              for j in range(self.ncol)] for i in range(self.nrow)]

def cell_neighbors(row: int, col: int, nrow: int, ncol: int
                   ) -> list[tuple[int,int]]:
    """Return the list of cells bordering a cell."""
    around = []
    if row != 0:
        around.append((row-1,col,))
    if row != nrow-1:
        around.append((row+1,col,))
    if col != 0:
        around.append((row,col-1,))
    if col != ncol-1:
        around.append((row,col+1,))

    return around

def find_free_neighbors(i: int, j: int,
                        grid_a: GridAnalysis
                        ) -> list[tuple[int,int]]:
    """Find empty cells in which regions may expand."""
    ngrid = grid_a.ngrid
    nrow = grid_a.nrow
    ncol = grid_a.ncol

    clue = ngrid[i][j]
    if clue is None:
        raise ValueError(f'empty cell ({i},{j}) sent to find_free_neighbors')

    region_neighbors: set[tuple[int,int]] = set(cell_neighbors(i, j, nrow, ncol))
    candidates = [(i1,j1,) for i1,j1 in region_neighbors if ngrid[i1][j1] is None]

    free_neighbors = list()
    for ic,jc in candidates:
        good = True
        cand_nbors = cell_neighbors(ic, jc, nrow, ncol)
        border = {(i,j,)}
        for ni,nj in cand_nbors:
            if ngrid[ni][nj] is not None and ngrid[ni][nj] == clue:
                border |= {(ni,nj,)}
                if len(border) + 1 > clue:
                    good = False
                    break
        if good:
            free_neighbors.append((ic,jc,))

    return free_neighbors

def mc_flow(grid_a: GridAnalysis,
            clues: dict[int,tuple[tuple[int,int],tuple[int,int]]],
            ) -> None:
    """Compute multi-commodity flow."""

    ngrid = grid_a.ngrid
    nrow = len(ngrid)
    ncol = len(ngrid[0])

    # vars[clue][i][j] <-> Commodity "clue" flows through cell (i,j).
    vars = dict()
    for clue in clues.keys():
        vars[clue] = [[Bool(f'v{clue}_{i}_{j}')
                       for j in range(ncol)] for i in range(nrow)]

    # vflow[i][j] <-> Some commodity flows through cell (i,j).
    vflow = [[Bool(f'f{i}_{j}') for j in range(ncol)] for i in range(nrow)]

    slv = SolverFor('QF_FD')

    # Define vflow variables in terms of the vars variables.
    slv.add([vflow[i][j] == Or([vars[clue][i][j] for clue in clues.keys()])
             for i in range(nrow) for j in range(ncol)])

    # Mutual exclusion among commodities.
    for i in range(nrow):
        for j in range(ncol):
            slv.add(PbLe([(vars[clue][i][j], 1) for clue in clues.keys()], 1))

    # Sources and sinks belong to the flows and they have
    # one neighbor with flow.  Obstructions have no flow.
    # For the other cells, we need to impose flow continuity.
    for i in range(nrow):
        for j in range(ncol):
            thisclue = ngrid[i][j]
            if thisclue is None:
                nbors = cell_neighbors(i, j, nrow, ncol)
                for clue in clues.keys():
                    slv.add(Implies(vars[clue][i][j],
                                    PbEq([(vars[clue][i1][j1], 1) for i1,j1 in nbors], 2)))
            else:
                for clue, terminals in clues.items():
                    if clue == thisclue:
                        nbors = cell_neighbors(i, j, nrow, ncol)
                        if (i,j,) in terminals:
                            slv.add(vars[clue][i][j])
                            slv.add(PbEq([(vars[clue][i1][j1], 1) for i1,j1 in nbors], 1))
                        else:
                            slv.add(Or(PbEq([(vars[clue][i1][j1], 1) for i1,j1 in nbors], 2),
                                       Not(vars[clue][i][j])))
                    else:
                        slv.add(Not(vars[clue][i][j]))

    # Cardinality.
    for clue in clues.keys():
        slv.add(PbLe([(vars[clue][i][j],1) for i in range(nrow) for j in range(ncol)], clue))

    # Room-for-growth constraints.
    seen = set()
    for i in range(nrow):
        for j in range(ncol):
            if (i,j,) in seen:
                continue
            thisclue = ngrid[i][j]
            if thisclue is not None and thisclue not in clues.keys():
                seen |= {(i,j,)}
                if 1 < thisclue:
                    free_nbors = find_free_neighbors(i, j, grid_a)
                    this_constraint = Or([Not(vflow[i1][j1]) for i1,j1 in free_nbors])
                    print(thisclue, this_constraint)
                    slv.add(this_constraint)

    cvars = [vflow[i][j] for i in range(nrow) for j in range(ncol)
             if ngrid[i][j] is None]
    result, conseq = slv.consequences(list(), cvars)

    if result == sat:
        print(conseq)
    else:
        print('mp_flow check returned:', result)

if __name__ == '__main__':

    _ = None
    grid = (( _, _, _, _, _,20,22,22,22,22,22, _),
            ( _, _, 5, _, _, _, _, _, _,20, _,22),
            ( _,20, _, 6, 6, 6, 6, 6, 6, _, 3,22),
            ( _, _, 3, _, _, _, _, _, _, 5, _,22),
            ( _, 9, 9, 9, 9,20,22, _, _, _, _, _),
            ( _, 9, 2, _, 9, _, _, _, _, 8, 4, _),
            ( _, 5, 9, 9, 9, _, _, _, _, 5, 8, _),
            ( _, _, _, _, _,20,22, _, _, _, _, _),
            ( _, _, 5, _, _, _, _, _, _, 3, _, _),
            ( _, 1, _, 5, _, _, _, _, 7, _, 7, _),
            ( _, _, 2, _, _, _, _, _, _, 3, _, _),
            ( _, _, _, _, _,20,22, _, _, _, _, _))

    grid_a = GridAnalysis(grid)

    mcf = {
        22: ((0, 6), (11, 6)),
        6: ((2, 3), (2, 8)),   # when this line is commented, no exception is raised
        9: ((5, 1), (6, 2)),
        7: ((9, 8), (9, 10))
           }
    mc_flow(grid_a,mcf)
